from xulbux import Console, Args, ArgResult
from xulbux import xx_console

from unittest.mock import MagicMock
from collections import namedtuple
import builtins
import pytest
import sys


@pytest.fixture
def mock_terminal_size(monkeypatch):
    TerminalSize = namedtuple('TerminalSize', ['columns', 'lines'])
    mock_get_terminal_size = lambda: TerminalSize(columns=80, lines=24)
    monkeypatch.setattr(xx_console._os, 'get_terminal_size', mock_get_terminal_size)


@pytest.fixture
def mock_formatcodes_print(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr(xx_console.FormatCodes, 'print', mock)
    return mock


@pytest.fixture
def mock_builtin_input(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr(builtins, 'input', mock)
    return mock


@pytest.fixture
def mock_prompt_toolkit(monkeypatch):
    mock = MagicMock(return_value="mocked multiline input")
    monkeypatch.setattr(xx_console._prompt_toolkit, 'prompt', mock)
    return mock


def test_console_user():
    user_output = Console.usr
    assert isinstance(user_output, str)
    assert user_output != ""


def test_console_width(mock_terminal_size):
    width_output = Console.w
    assert isinstance(width_output, int)
    assert width_output == 80


def test_console_height(mock_terminal_size):
    height_output = Console.h
    assert isinstance(height_output, int)
    assert height_output == 24


def test_console_size(mock_terminal_size):
    size_output = Console.wh
    assert isinstance(size_output, tuple)
    assert len(size_output) == 2
    assert size_output[0] == 80
    assert size_output[1] == 24


@pytest.mark.parametrize(
    "argv, find_args, expected_args_dict", [
        # --- CASES WITHOUT SPACES (allow_spaces=False) ---
        # NO ARGS PROVIDED
        (["script.py"], {"file": ["-f"], "debug": ["-d"]
                         }, {"file": {"exists": False, "value": None}, "debug": {"exists": False, "value": None}}),
        # SIMPLE FLAG
        (["script.py", "-d"], {"file": ["-f"], "debug": ["-d"]
                               }, {"file": {"exists": False, "value": None}, "debug": {"exists": True, "value": True}}),
        # FLAG WITH VALUE
        (["script.py", "-f", "test.txt"], {"file": ["-f"], "debug": ["-d"]},
         {"file": {"exists": True, "value": "test.txt"}, "debug": {"exists": False, "value": None}}),
        # LONG FLAGS WITH VALUE AND FLAG
        (["script.py", "--file", "path/to/file", "--debug"], {"file": ["-f", "--file"], "debug": ["-d", "--debug"]},
         {"file": {"exists": True, "value": "path/to/file"}, "debug": {"exists": True, "value": True}}),
        # VALUE WITH SPACE (IGNORED DUE TO allow_spaces=False)
        (["script.py", "-f", "file with spaces"], {"file": ["-f"]}, {"file": {"exists": True, "value": "file with spaces"}}),
        # UNKNOWN ARG
        (["script.py", "-x"], {"file": ["-f"]}, {"file": {"exists": False, "value": None}}),
        # TWO FLAGS
        (["script.py", "-f", "-d"], {"file": ["-f"], "debug": ["-d"]
                                     }, {"file": {"exists": True, "value": True}, "debug": {"exists": True, "value": True}}),
        # NUMERIC VALUE
        (["script.py", "-n", "123"], {"num": ["-n"]}, {"num": {"exists": True, "value": 123}}),
        # BOOLEAN VALUE (True)
        (["script.py", "-b", "true"], {"bool_arg": ["-b"]}, {"bool_arg": {"exists": True, "value": True}}),
        # BOOLEAN VALUE (False)
        (["script.py", "-b", "False"], {"bool_arg": ["-b"]}, {"bool_arg": {"exists": True, "value": False}}),

        # --- CASES WITH DEFAULTS (dict FORMAT, allow_spaces=False) ---
        # DEFAULT USED (string)
        (["script.py"], {"output": {"flags": ["-o"], "default": "out.txt"}, "verbose": ["-v"]
                         }, {"output": {"exists": False, "value": "out.txt"}, "verbose": {"exists": False, "value": None}}),
        # VALUE OVERRIDES DEFAULT (string)
        (["script.py", "-o", "my_file.log"], {"output": {"flags": ["-o"], "default": "out.txt"}, "verbose": ["-v"]},
         {"output": {"exists": True, "value": "my_file.log"}, "verbose": {"exists": False, "value": None}}),
        # FLAG PRESENCE OVERRIDES DEFAULT (string -> True)
        (["script.py", "-o"], {"output": {"flags": ["-o"], "default": "out.txt"}, "verbose": ["-v"]
                               }, {"output": {"exists": True, "value": True}, "verbose": {"exists": False, "value": None}}),
        # FLAG PRESENCE OVERRIDES DEFAULT (False -> True)
        (["script.py", "-v"], {
            "output": {"flags": ["-o"], "default": "out.txt"}, "verbose": {"flags": ["-v"], "default": False}
        }, {"output": {"exists": False, "value": "out.txt"}, "verbose": {"exists": True, "value": True}}),
        # DEFAULT USED (int)
        (["script.py"], {"mode": {"flags": ["-m"], "default": 1}}, {"mode": {"exists": False, "value": 1}}),
        # VALUE OVERRIDES DEFAULT (int)
        (["script.py", "-m", "2"], {"mode": {"flags": ["-m"], "default": 1}}, {"mode": {"exists": True, "value": 2}}),
        # FLAG PRESENCE OVERRIDES DEFAULT (int -> True)
        (["script.py", "-m"], {"mode": {"flags": ["-m"], "default": 1}}, {"mode": {"exists": True, "value": True}}),

        # --- MIXED list/tuple AND dict FORMATS (allow_spaces=False) ---
        # DICT VALUE PROVIDED, LIST NOT PROVIDED
        (["script.py", "--config", "dev.cfg"], {"config": {"flags": ["-c", "--config"], "default": "prod.cfg"}, "log": ["-l"]},
         {"config": {"exists": True, "value": "dev.cfg"}, "log": {"exists": False, "value": None}}),
        # LIST FLAG PROVIDED, DICT NOT PROVIDED (USES DEFAULT)
        (["script.py", "-l"], {"config": {"flags": ["-c", "--config"], "default": "prod.cfg"}, "log": ["-l"]
                               }, {"config": {"exists": False, "value": "prod.cfg"}, "log": {"exists": True, "value": True}}),
    ]
)
def test_get_args_no_spaces(monkeypatch, argv, find_args, expected_args_dict):
    monkeypatch.setattr(sys, 'argv', argv)
    args_result = Console.get_args(find_args, allow_spaces=False)
    assert isinstance(args_result, Args)
    assert args_result.dict() == expected_args_dict
    for key, expected in expected_args_dict.items():
        assert (key in args_result) is True
        assert isinstance(args_result[key], ArgResult)
        assert args_result[key].exists == expected["exists"]  # type: ignore[access]
        assert args_result[key].value == expected["value"]  # type: ignore[access]
        assert bool(args_result[key]) == expected["exists"]
    assert list(args_result.keys()) == list(expected_args_dict.keys())
    assert [v.exists for v in args_result.values()] == [d["exists"] for d in expected_args_dict.values()]
    assert [v.value for v in args_result.values()] == [d["value"] for d in expected_args_dict.values()]
    assert len(args_result) == len(expected_args_dict)


@pytest.mark.parametrize(
    "argv, find_args, expected_args_dict", [
        # --- CASES WITH SPACES (allow_spaces=True) ---
        # SIMPLE VALUE WITH SPACES
        (["script.py", "-f", "file with spaces", "-d"], {"file": ["-f"], "debug": ["-d"]},
         {"file": {"exists": True, "value": "file with spaces"}, "debug": {"exists": True, "value": True}}),
        # LONG VALUE WITH SPACES
        (["script.py", "--message", "Hello", "world", "how", "are", "you"
          ], {"message": ["--message"]}, {"message": {"exists": True, "value": "Hello world how are you"}}),
        # VALUE WITH SPACES FOLLOWED BY ANOTHER FLAG
        (["script.py", "-m", "this is", "a message", "--flag"], {"message": ["-m"], "flag": ["--flag"]},
         {"message": {"exists": True, "value": "this is a message"}, "flag": {"exists": True, "value": True}}),
        # VALUE WITH SPACES AT THE END
        (["script.py", "-m", "end", "of", "args"], {"message": ["-m"]}, {"message": {"exists": True, "value": "end of args"}}),

        # --- CASES WITH DEFAULTS (dict FORMAT, allow_spaces=True) ---
        # VALUE WITH SPACE OVERRIDES DEFAULT
        (["script.py", "--msg", "Default message"], {"msg": {"flags": ["--msg"], "default": "No message"}, "other": ["-o"]},
         {"msg": {"exists": True, "value": "Default message"}, "other": {"exists": False, "value": None}}),
        # DEFAULT USED WHEN OTHER FLAG PRESENT
        (["script.py", "-o"], {"msg": {"flags": ["--msg"], "default": "No message"}, "other": ["-o"]
                               }, {"msg": {"exists": False, "value": "No message"}, "other": {"exists": True, "value": True}}),
        # FLAG PRESENCE OVERRIDES DEFAULT (str -> True)
        (["script.py", "--msg"], {"msg": {"flags": ["--msg"], "default": "No message"}, "other": ["-o"]
                                  }, {"msg": {"exists": True, "value": True}, "other": {"exists": False, "value": None}}),

        # --- MIXED FORMATS WITH SPACES (allow_spaces=True) ---
        # LIST VALUE WITH SPACES, dict VALUE PROVIDED
        (["script.py", "-f", "input file name", "--mode", "test"], {
            "file": ["-f"], "mode": {"flags": ["--mode"], "default": "prod"}
        }, {"file": {"exists": True, "value": "input file name"}, "mode": {"exists": True, "value": "test"}}),
        # LIST VALUE WITH SPACES, dict NOT PROVIDED (USES DEFAULT)
        (["script.py", "-f", "another file"], {"file": ["-f"], "mode": {"flags": ["--mode"], "default": "prod"}},
         {"file": {"exists": True, "value": "another file"}, "mode": {"exists": False, "value": "prod"}}),
    ]
)
def test_get_args_with_spaces(monkeypatch, argv, find_args, expected_args_dict):
    monkeypatch.setattr(sys, 'argv', argv)
    args_result = Console.get_args(find_args, allow_spaces=True)
    assert isinstance(args_result, Args)
    assert args_result.dict() == expected_args_dict


def test_get_args_invalid_alias():
    with pytest.raises(TypeError, match="Argument alias 'invalid-alias' is invalid."):
        Args(**{"invalid-alias": {"exists": False, "value": None}})

    with pytest.raises(TypeError, match="Argument alias '123start' is invalid."):
        Args(**{"123start": {"exists": False, "value": None}})


def test_get_args_invalid_config():
    with pytest.raises(TypeError, match="Invalid configuration type for alias 'bad_config'. Must be a list, tuple, or dict."):
        Console.get_args({"bad_config": 123})  # type: ignore[assignment]

    with pytest.raises(ValueError,
                       match="Invalid configuration for alias 'missing_flags'. Dictionary must contain a 'flags' key."):
        Console.get_args({"missing_flags": {"default": "value"}})

    with pytest.raises(ValueError, match="Invalid 'flags' for alias 'bad_flags'. Must be a list or tuple."):
        Console.get_args({"bad_flags": {"flags": "not-a-list"}})


def test_get_args_duplicate_flag():
    with pytest.raises(ValueError, match="Duplicate flag '-f' found. It's assigned to both 'file1' and 'file2'."):
        Console.get_args({"file1": ["-f", "--file1"], "file2": {"flags": ["-f", "--file2"]}})

    with pytest.raises(ValueError, match="Duplicate flag '--long' found. It's assigned to both 'arg1' and 'arg2'."):
        Console.get_args({"arg1": {"flags": ["--long"]}, "arg2": ("-a", "--long")})


def test_multiline_input(mock_prompt_toolkit, mock_formatcodes_print):
    expected_input = "mocked multiline input"
    result = Console.multiline_input("Enter text:", show_keybindings=True, default_color="#BCA")

    assert result == expected_input
    assert mock_formatcodes_print.call_count == 3
    prompt_call = mock_formatcodes_print.call_args_list[0]
    keybind_call = mock_formatcodes_print.call_args_list[1]
    reset_call = mock_formatcodes_print.call_args_list[2]

    assert prompt_call.args == ("Enter text:", )
    assert prompt_call.kwargs == {'default_color': '#BCA'}

    assert "[dim][[b](CTRL+D)[dim] : end of input][_dim]" in keybind_call.args[0]

    assert reset_call.args == ('[_]', )
    assert reset_call.kwargs == {'end': ''}

    mock_prompt_toolkit.assert_called_once()
    pt_args, pt_kwargs = mock_prompt_toolkit.call_args
    assert pt_args == (" ⤷ ", )
    assert pt_kwargs.get('multiline') is True
    assert pt_kwargs.get('wrap_lines') is True
    assert 'key_bindings' in pt_kwargs


def test_multiline_input_no_bindings(mock_prompt_toolkit, mock_formatcodes_print):
    Console.multiline_input("Enter text:", show_keybindings=False, end="DONE")

    assert mock_formatcodes_print.call_count == 2
    prompt_call = mock_formatcodes_print.call_args_list[0]
    reset_call = mock_formatcodes_print.call_args_list[1]

    assert prompt_call.args == ("Enter text:", )
    assert reset_call.args == ('[_]', )
    assert reset_call.kwargs == {'end': 'DONE'}

    mock_prompt_toolkit.assert_called_once()



from ._consts_ import COLOR
from . import __version__
from .xx_format_codes import FormatCodes
from .xx_console import Console


def help_command():
    """Show some info about the library, with a brief explanation of how to use it."""
    color = {
        "class": COLOR.tangerine,
        "const": COLOR.red,
        "func": COLOR.cyan,
        "import": COLOR.neongreen,
        "lib": COLOR.orange,
        "punctuators": COLOR.darkgray,
        "code_border": COLOR.gray,
    }
    FormatCodes.print(
        rf"""  [_|b|#7075FF]               __  __
  [b|#7075FF]  _  __ __  __/ / / /_  __  ___  __
  [b|#7075FF] | |/ // / / / / / __ \/ / / | |/ /
  [b|#7075FF] > , </ /_/ / /_/ /_/ / /_/ /> , <
  [b|#7075FF]/_/|_|\____/\__/\____/\____//_/|_|  [*|BG:{COLOR.gray}|#000] v[b]{__version__} [*]

  [i|{COLOR.coral}]A TON OF COOL FUNCTIONS, YOU NEED![*]

  [b|#FCFCFF]Usage:[*]
  [dim|{color['code_border']}](╭────────────────────────────────────────────────────╮)
  [dim|{color['code_border']}](│) [{color['punctuators']}]# CONSTANTS[*]                                        [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](│) [{color['import']}]from [{color['lib']}]xulbux [{color['import']}]import [{color['const']}]COLOR[{color['punctuators']}], [{color['const']}]CHARS[{color['punctuators']}], [{color['const']}]ANSI[*]              [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](│) [{color['punctuators']}]# Classes[*]                                          [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](│) [{color['import']}]from [{color['lib']}]xulbux [{color['import']}]import [{color['class']}]Code[{color['punctuators']}], [{color['class']}]Color[{color['punctuators']}], [{color['class']}]Console[{color['punctuators']}], ...[*]       [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](│) [{color['punctuators']}]# types[*]                                            [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](│) [{color['import']}]from [{color['lib']}]xulbux [{color['import']}]import [{color['func']}]rgba[{color['punctuators']}], [{color['func']}]hsla[{color['punctuators']}], [{color['func']}]hexa[*]                [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](╰────────────────────────────────────────────────────╯)
  [b|#FCFCFF]Documentation:[*]
  [dim|{color['code_border']}](╭────────────────────────────────────────────────────╮)
  [dim|{color['code_border']}](│) [#DADADD]For more information see the GitHub page.          [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](│) [u|#8085FF](https://github.com/XulbuX/PythonLibraryXulbuX/wiki) [dim|{color['code_border']}](│)
  [dim|{color['code_border']}](╰────────────────────────────────────────────────────╯)
  [_]
  [dim](Press any key to exit...)
  """,
        default_color=COLOR.text
    )
    Console.pause_exit(pause=True)



from .xx_data import Data
from .xx_file import File
from .xx_path import Path

from typing import Any
import json as _json


class Json:

    @staticmethod
    def read(
        json_file: str,
        comment_start: str = ">>",
        comment_end: str = "<<",
        return_original: bool = False,
    ) -> dict | tuple[dict, dict]:
        """Read JSON files, ignoring comments.\n
        ------------------------------------------------------------------
        If only `comment_start` is found at the beginning of an item,
        the whole item is counted as a comment and therefore ignored.
        If `comment_start` and `comment_end` are found inside an item,
        the the section from `comment_start` to `comment_end` is ignored.
        If `return_original` is true, the original JSON is returned
        additionally. (returns: `[processed_json, original_json]`)"""
        if not json_file.endswith(".json"):
            json_file += ".json"
        file_path = Path.extend_or_make(json_file, prefer_script_dir=True)
        if file_path is None:
            raise FileNotFoundError(f"Could not find JSON file: {json_file}")
        with open(file_path, "r") as f:
            content = f.read()
        try:
            data = _json.loads(content)
        except _json.JSONDecodeError as e:
            raise ValueError(f"Error parsing JSON in '{file_path}':  {str(e)}")
        processed_data = dict(Data.remove_comments(data, comment_start, comment_end))
        if not processed_data:
            raise ValueError(f"The JSON file '{file_path}' is empty or contains only comments.")
        return (processed_data, data) if return_original else processed_data

    @staticmethod
    def create(
        json_file: str,
        data: dict,
        indent: int = 2,
        compactness: int = 1,
        force: bool = False,
    ) -> str:
        """Create a nicely formatted JSON file from a dictionary.\n
        ----------------------------------------------------------------------
        The `indent` is the amount of spaces to use for indentation.\n
        The `compactness` can be `0`, `1` or `2` and indicates how compact
        the data should be formatted (see `Data.to_str()`).\n
        The function will throw a `FileExistsError` if a file with the same
        name already exists and a `SameContentFileExistsError` if a file with
        the same name and content already exists.
        To always overwrite the file, set the `force` parameter to `True`."""
        if not json_file.endswith(".json"):
            json_file += ".json"
        file_path = Path.extend_or_make(json_file, prefer_script_dir=True)
        File.create(
            file=file_path,
            content=Data.to_str(data, indent, compactness, as_json=True),
            force=force,
        )
        return file_path

    @staticmethod
    def update(
        json_file: str,
        update_values: dict[str, Any],
        comment_start: str = ">>",
        comment_end: str = "<<",
        path_sep: str = "->",
    ) -> None:
        """Update single/multiple values inside JSON files, without needing to know the rest of the data.\n
        ----------------------------------------------------------------------------------------------------
        The `update_values` parameter is a dictionary, where the keys are the paths to the data to update,
        and the values are the new values to set.\n
        Example: For this JSON data:
        ```python
        {
            "healthy": {
                "fruit": ["apples", "bananas", "oranges"],
                "vegetables": ["carrots", "broccoli", "celery"]
            }
        }
        ```
        ... the `update_values` dictionary could look like this:
        ```python
        {
            # CHANGE VALUE "apples" TO "strawberries"
            "healthy->fruit->0": "strawberries",
            # CHANGE VALUE UNDER KEY "vegetables" TO [1, 2, 3]
            "healthy->vegetables": [1, 2, 3]
        }
        ```
        In this example, if you want to change the value of `"apples"`, you can use `healthy->fruit->apples`
        as the value-path. If you don't know that the first list item is `"apples"`, you can use the items
        list index inside the value-path, so `healthy->fruit->0`.\n
        ⇾ If the given value-path doesn't exist, it will be created.\n
        -----------------------------------------------------------------------------------------------------
        If only `comment_start` is found at the beginning of an item, the whole item is counted as a comment
        and therefore completely ignored. If `comment_start` and `comment_end` are found inside an item, the
        section from `comment_start` to `comment_end` is counted as a comment and ignored."""
        processed_data, data = Json.read(json_file, comment_start, comment_end, return_original=True)

        def create_nested_path(data_obj: dict, path_keys: list[str], value: Any) -> dict:
            current = data_obj
            last_idx = len(path_keys) - 1
            for i, key in enumerate(path_keys):
                if i == last_idx:
                    if isinstance(current, dict):
                        current[key] = value
                    elif isinstance(current, list) and key.isdigit():
                        idx = int(key)
                        while len(current) <= idx:
                            current.append(None)
                        current[idx] = value
                    else:
                        raise TypeError(f"Cannot set key '{key}' on {type(current).__name__}")
                else:
                    next_key = path_keys[i + 1]
                    if isinstance(current, dict):
                        if key not in current:
                            current[key] = [] if next_key.isdigit() else {}
                        current = current[key]
                    elif isinstance(current, list) and key.isdigit():
                        idx = int(key)
                        while len(current) <= idx:
                            current.append(None)
                        if current[idx] is None:
                            current[idx] = [] if next_key.isdigit() else {}
                        current = current[idx]
                    else:
                        raise TypeError(f"Cannot navigate through {type(current).__name__}")
            return data_obj

        update = {}
        for value_path, new_value in update_values.items():
            try:
                path_id = Data.get_path_id(
                    data=processed_data,
                    value_paths=value_path,
                    path_sep=path_sep,
                )
                if path_id is not None:
                    update[path_id] = new_value
                else:
                    keys = value_path.split(path_sep)
                    keys = value_path.split(path_sep)
                    data = create_nested_path(data, keys, new_value)
            except Exception:
                keys = value_path.split(path_sep)
                data = create_nested_path(data, keys, new_value)
        if "update" in locals() and update:
            data = Data.set_value_by_path_id(data, update)
        Json.create(json_file=json_file, data=dict(data), force=True)



from dataclasses import dataclass
from typing import TypeAlias


FormattableString: TypeAlias = str
"""A `str` object that is made to be formatted with the `.format()` method."""


@dataclass
class COLOR:
    """Color presets used in the `xulbux` library."""

    text = "#A5D6FF"
    """The default text color used in the `xulbux` library."""

    white = "#F1F2FF"
    lightgray = "#B6B7C0"
    gray = "#7B7C8D"
    darkgray = "#67686C"
    black = "#202125"
    red = "#FF606A"
    coral = "#FF7069"
    orange = "#FF876A"
    tangerine = "#FF9962"
    gold = "#FFAF60"
    yellow = "#FFD260"
    lime = "#C9F16E"
    green = "#7EE787"
    neongreen = "#4CFF85"
    teal = "#50EAAF"
    cyan = "#3EDEE6"
    ice = "#77DBEF"
    lightblue = "#60AAFF"
    blue = "#8085FF"
    lavender = "#9B7DFF"
    purple = "#AD68FF"
    magenta = "#C860FF"
    pink = "#F162EF"
    rose = "#FF609F"


class _AllTextCharacters:
    pass


@dataclass
class CHARS:
    """Text character sets."""

    all = _AllTextCharacters
    """Code to signal that all characters are allowed."""

    digits: str = "0123456789"
    """Digits: `0`-`9`"""
    float_digits: str = digits + "."
    """Digits: `0`-`9` with decimal point `.`"""
    hex_digits: str = digits + "#abcdefABCDEF"
    """Digits: `0`-`9` Letters: `a`-`f` `A`-`F` and a hashtag `#`"""

    lowercase: str = "abcdefghijklmnopqrstuvwxyz"
    """Lowercase letters `a`-`z`"""
    lowercase_extended: str = lowercase + "äëïöüÿàèìòùáéíóúýâêîôûãñõåæç"
    """Lowercase letters `a`-`z` with all lowercase diacritic letters."""
    uppercase: str = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    """Uppercase letters `A`-`Z`"""
    uppercase_extended: str = uppercase + "ÄËÏÖÜÀÈÌÒÙÁÉÍÓÚÝÂÊÎÔÛÃÑÕÅÆÇß"
    """Uppercase letters `A`-`Z` with all uppercase diacritic letters."""

    letters: str = lowercase + uppercase
    """Lowercase and uppercase letters `a`-`z` and `A`-`Z`"""
    letters_extended: str = lowercase_extended + uppercase_extended
    """Lowercase and uppercase letters `a`-`z` `A`-`Z` and all diacritic letters."""

    special_ascii: str = " !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"
    """All ASCII special characters."""
    special_ascii_extended: str = special_ascii + "ø£Ø×ƒªº¿®¬½¼¡«»░▒▓│┤©╣║╗╝¢¥┐└┴┬├─┼╚╔╩╦╠═╬¤ðÐı┘┌█▄¦▀µþÞ¯´≡­±‗¾¶§÷¸°¨·¹³²■ "
    """All ASCII special characters with the extended ASCII special characters."""
    standard_ascii: str = special_ascii + digits + letters
    """All standard ASCII characters."""
    full_ascii: str = special_ascii_extended + digits + letters_extended
    """All characters in the ASCII table."""


class ANSI:
    """Constants and class-methods for use of ANSI escape codes."""

    escaped_char: str = "\\x1b"
    """The printable ANSI escape character."""
    CHAR = char = "\x1b"
    """The ANSI escape character."""
    START = start = "["
    """The start of an ANSI escape sequence."""
    SEP = sep = ";"
    """The separator between ANSI escape sequence parts."""
    END = end = "m"
    """The end of an ANSI escape sequence."""
    default_color_modifiers: dict[str, str] = {"lighten": "+l", "darken": "-d"}
    """Characters to modify the lightness of the default color with."""

    @classmethod
    def seq(cls, parts: int = 1) -> FormattableString:
        """Generate an ANSI sequence with `parts` amount of placeholders."""
        return cls.CHAR + cls.START + cls.SEP.join(["{}" for _ in range(parts)]) + cls.END

    seq_color: FormattableString = CHAR + START + "38" + SEP + "2" + SEP + "{}" + SEP + "{}" + SEP + "{}" + END
    """The ANSI escape sequence for setting the text RGB color."""
    seq_bg_color: FormattableString = CHAR + START + "48" + SEP + "2" + SEP + "{}" + SEP + "{}" + SEP + "{}" + END
    """The ANSI escape sequence for setting the background RGB color."""

    color_map: list[str] = [
        ########### DEFAULT CONSOLE COLOR NAMES ############
        "black",
        "red",
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
    ]
    """The console default color names."""

    codes_map: dict[str | tuple[str, ...], int] = {
        ################# SPECIFIC RESETS ##################
        "_": 0,
        ("_bold", "_b"): 22,
        ("_dim", "_d"): 22,
        ("_italic", "_i"): 23,
        ("_underline", "_u"): 24,
        ("_double-underline", "_du"): 24,
        ("_inverse", "_invert", "_in"): 27,
        ("_hidden", "_hide", "_h"): 28,
        ("_strikethrough", "_s"): 29,
        ("_color", "_c"): 39,
        ("_background", "_bg"): 49,
        ################### TEXT STYLES ####################
        ("bold", "b"): 1,
        ("dim", "d"): 2,
        ("italic", "i"): 3,
        ("underline", "u"): 4,
        ("inverse", "invert", "in"): 7,
        ("hidden", "hide", "h"): 8,
        ("strikethrough", "s"): 9,
        ("double-underline", "du"): 21,
        ################## DEFAULT COLORS ##################
        "black": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "magenta": 35,
        "cyan": 36,
        "white": 37,
        ############## BRIGHT DEFAULT COLORS ###############
        "br:black": 90,
        "br:red": 91,
        "br:green": 92,
        "br:yellow": 93,
        "br:blue": 94,
        "br:magenta": 95,
        "br:cyan": 96,
        "br:white": 97,
        ############ DEFAULT BACKGROUND COLORS #############
        "bg:black": 40,
        "bg:red": 41,
        "bg:green": 42,
        "bg:yellow": 43,
        "bg:blue": 44,
        "bg:magenta": 45,
        "bg:cyan": 46,
        "bg:white": 47,
        ######### BRIGHT DEFAULT BACKGROUND COLORS #########
        "bg:br:black": 100,
        "bg:br:red": 101,
        "bg:br:green": 102,
        "bg:br:yellow": 103,
        "bg:br:blue": 104,
        "bg:br:magenta": 105,
        "bg:br:cyan": 106,
        "bg:br:white": 107,
    }
    """All custom format keys and their corresponding ANSI format number codes."""



from xulbux._consts_ import ANSI
from xulbux import FormatCodes


black = ANSI.seq_color.format(0, 0, 0)
bg_red = f"{ANSI.char}{ANSI.start}{ANSI.codes_map['bg:red']}{ANSI.end}"
default = ANSI.seq_color.format(255, 255, 255)
orange = ANSI.seq_color.format(255, 136, 119)

bold = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('bold', 'b')]}{ANSI.end}"
invert = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('inverse', 'invert', 'in')]}{ANSI.end}"
italic = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('italic', 'i')]}{ANSI.end}"
underline = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('underline', 'u')]}{ANSI.end}"

reset = f"{ANSI.char}{ANSI.start}{ANSI.codes_map['_']}{ANSI.end}"
reset_bg = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_background', '_bg')]}{ANSI.end}"
reset_bold = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_bold', '_b')]}{ANSI.end}"
reset_color = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_color', '_c')]}{ANSI.end}"
reset_italic = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_italic', '_i')]}{ANSI.end}"
reset_invert = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_inverse', '_invert', '_in')]}{ANSI.end}"
reset_underline = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_underline', '_u')]}{ANSI.end}"


def test_to_ansi():
    assert (
        FormatCodes.to_ansi("[b|#000|bg:red](He[in](l)lo) [[i|u|#F87](world)][default]![_]",
                            default_color="#FFF") == f"{default}{bold}{black}{bg_red}" + "He" + invert + "l" + reset_invert
        + "lo" + f"{reset_bold}{reset_color}{reset_bg}" + " [" + f"{italic}{underline}{orange}" + "world"
        + f"{reset_italic}{reset_underline}{reset_color}" + "]" + default + "!" + reset
    )


def test_escape_ansi():
    ansi_string = f"{bold}Hello {orange}World!{reset}"
    escaped_string = ansi_string.replace(ANSI.char, ANSI.escaped_char)
    assert FormatCodes.escape_ansi(ansi_string) == escaped_string


def test_remove_ansi():
    ansi_string = f"{bold}Hello {orange}World!{reset}"
    clean_string = "Hello World!"
    assert FormatCodes.remove_ansi(ansi_string) == clean_string


def test_remove_ansi_with_removals():
    ansi_string = f"{bold}Hello {orange}World!{reset}"
    clean_string = "Hello World!"
    removals = ((0, bold), (6, orange), (12, reset))
    assert FormatCodes.remove_ansi(ansi_string, get_removals=True) == (clean_string, removals)


def test_remove_formatting():
    format_string = "[b](Hello [#F87](World!))"
    clean_string = "Hello World!"
    assert FormatCodes.remove_formatting(format_string) == clean_string


def test_remove_formatting_with_removals():
    format_string = "[b](Hello [#F87](World!))"
    clean_string = "Hello World!"
    removals = ((0, bold), (6, orange), (12, reset_color), (12, reset_bold))
    assert FormatCodes.remove_formatting(format_string, get_removals=True) == (clean_string, removals)



from ._consts_ import COLOR
from .xx_format_codes import FormatCodes
from .xx_string import String

from typing import TypeAlias, Optional, Union, Any
import base64 as _base64
import math as _math
import re as _re


DataStructure: TypeAlias = Union[list, tuple, set, frozenset, dict]
IndexIterable: TypeAlias = Union[list, tuple, set, frozenset]


class Data:

    @staticmethod
    def serialize_bytes(data: bytes | bytearray) -> dict[str, str]:
        """Converts bytes or bytearray to a JSON-compatible format (dictionary) with explicit keys."""
        if isinstance(data, (bytes, bytearray)):
            key = "bytearray" if isinstance(data, bytearray) else "bytes"
            try:
                return {key: data.decode("utf-8"), "encoding": "utf-8"}
            except UnicodeDecodeError:
                pass
            return {key: _base64.b64encode(data).decode("utf-8"), "encoding": "base64"}
        raise TypeError("Unsupported data type")

    @staticmethod
    def deserialize_bytes(obj: dict[str, str]) -> bytes | bytearray:
        """Converts a JSON-compatible bytes/bytearray format (dictionary) back to its original type."""
        for key in ("bytes", "bytearray"):
            if key in obj and "encoding" in obj:
                if obj["encoding"] == "utf-8":
                    data = obj[key].encode("utf-8")
                elif obj["encoding"] == "base64":
                    data = _base64.b64decode(obj[key].encode("utf-8"))
                else:
                    raise ValueError("Unknown encoding method")
                return bytearray(data) if key == "bytearray" else data
        raise ValueError("Invalid serialized data")

    @staticmethod
    def chars_count(data: DataStructure) -> int:
        """The sum of all the characters amount including the keys in dictionaries."""
        chars_count = 0
        if isinstance(data, dict):
            for k, v in data.items():
                chars_count += len(str(k)) + (Data.chars_count(v) if isinstance(v, DataStructure) else len(str(v)))
        elif isinstance(data, IndexIterable):
            for item in data:
                chars_count += Data.chars_count(item) if isinstance(item, DataStructure) else len(str(item))
        return chars_count

    @staticmethod
    def strip(data: DataStructure) -> DataStructure:
        """Removes leading and trailing whitespaces from the data structure's items."""
        if isinstance(data, dict):
            return {k.strip(): Data.strip(v) if isinstance(v, DataStructure) else v.strip() for k, v in data.items()}
        if isinstance(data, IndexIterable):
            return type(data)(Data.strip(item) if isinstance(item, DataStructure) else item.strip() for item in data)
        return data

    @staticmethod
    def remove_empty_items(data: DataStructure, spaces_are_empty: bool = False) -> DataStructure:
        """Removes empty items from the data structure.
        If `spaces_are_empty` is true, it will count items with only spaces as empty."""
        if isinstance(data, dict):
            return {
                k: (v if not isinstance(v, DataStructure) else Data.remove_empty_items(v, spaces_are_empty))
                for k, v in data.items() if not String.is_empty(v, spaces_are_empty)
            }
        if isinstance(data, IndexIterable):
            return type(data)(
                item for item in
                ((item if not isinstance(item, DataStructure) else Data.remove_empty_items(item, spaces_are_empty))
                 for item in data if not String.is_empty(item, spaces_are_empty))
                if item not in ([], (), {}, set(), frozenset())
            )
        return data

    @staticmethod
    def remove_duplicates(data: DataStructure) -> DataStructure:
        """Removes all duplicates from the data structure."""
        if isinstance(data, dict):
            return {k: Data.remove_duplicates(v) if isinstance(v, DataStructure) else v for k, v in data.items()}
        if isinstance(data, (list, tuple)):
            result = []
            for item in data:
                processed_item = Data.remove_duplicates(item) if isinstance(item, DataStructure) else item
                is_duplicate = False
                for existing_item in result:
                    if processed_item == existing_item:
                        is_duplicate = True
                        break
                if not is_duplicate:
                    result.append(processed_item)
            return type(data)(result)
        if isinstance(data, (set, frozenset)):
            processed_elements = set()
            for item in data:
                processed_item = Data.remove_duplicates(item) if isinstance(item, DataStructure) else item
                processed_elements.add(processed_item)
            return type(data)(processed_elements)
        return data

    @staticmethod
    def remove_comments(
        data: DataStructure,
        comment_start: str = ">>",
        comment_end: str = "<<",
        comment_sep: str = "",
    ) -> DataStructure:
        """Remove comments from a list, tuple or dictionary.\n
        ----------------------------------------------------------------------------------------------------------------------
        - The `data` parameter is your list, tuple or dictionary, where the comments should get removed from.
        - The `comment_start` parameter is the string that marks the start of a comment inside `data`. (default: `>>`)
        - The `comment_end` parameter is the string that marks the end of a comment inside `data`. (default: `<<`)
        - The `comment_sep` parameter is a string with which a comment will be replaced, if it is in the middle of a value.\n
        ----------------------------------------------------------------------------------------------------------------------
        Examples:
        ```python
        data = {
            "key1": [
                ">> COMMENT IN THE BEGINNING OF THE STRING <<  value1",
                "value2  >> COMMENT IN THE END OF THE STRING",
                "val>> COMMENT IN THE MIDDLE OF THE STRING <<ue3",
                ">> FULL VALUE IS A COMMENT  value4"
            ],
            ">> FULL KEY + ALL ITS VALUES ARE A COMMENT  key2": [
                "value",
                "value",
                "value"
            ],
            "key3": ">> ALL THE KEYS VALUES ARE COMMENTS  value"
        }

        processed_data = Data.remove_comments(
            data,
            comment_start=">>",
            comment_end="<<",
            comment_sep="__"
        )
        ```\n
        ----------------------------------------------------------------------------------------------------------------------
        For this example, `processed_data` will be:
        ```python
        {
            "key1": [
                "value1",
                "value2",
                "val__ue3"
            ],
            "key3": None
        }
        ```\n
        - For `key1`, all the comments will just be removed, except at `value3` and `value4`:
          - `value3` The comment is removed and the parts left and right are joined through `comment_sep`.
          - `value4` The whole value is removed, since the whole value was a comment.
        - For `key2`, the key, including its whole values will be removed.
        - For `key3`, since all its values are just comments, the key will still exist, but with a value of `None`."""
        if comment_end:
            pattern = _re.compile(
                rf"^((?:(?!{_re.escape(comment_start)}).)*){_re.escape(comment_start)}(?:(?:(?!{_re.escape(comment_end)}).)*)(?:{_re.escape(comment_end)})?(.*?)$"
            )

        def process_string(s: str) -> Optional[str]:
            if comment_end:
                match = pattern.match(s)  # type: ignore[unbound]
                if match:
                    start, end = match.group(1).strip(), match.group(2).strip()
                    return f"{start}{comment_sep if start and end else ''}{end}" or None
                return s.strip() or None
            else:
                return None if s.lstrip().startswith(comment_start) else s.strip() or None

        def process_item(item: Any) -> Any:
            if isinstance(item, dict):
                return {
                    k: v
                    for k, v in ((process_item(key), process_item(value)) for key, value in item.items()) if k is not None
                }
            if isinstance(item, IndexIterable):
                processed = (v for v in map(process_item, item) if v is not None)
                return type(item)(processed)
            if isinstance(item, str):
                return process_string(item)
            return item

        return process_item(data)

    @staticmethod
    def is_equal(
        data1: DataStructure,
        data2: DataStructure,
        ignore_paths: str | list[str] = "",
        path_sep: str = "->",
        comment_start: str = ">>",
        comment_end: str = "<<",
    ) -> bool:
        """Compares two structures and returns `True` if they are equal and `False` otherwise.\n
        ⇾ Will not detect, if a key-name has changed, only if removed or added.\n
        --------------------------------------------------------------------------------------------
        Ignores the specified (found) key/s or item/s from `ignore_paths`. Comments are not ignored
        when comparing. `comment_start` and `comment_end` are only used to correctly recognize the
        keys in the `ignore_paths`.\n
        --------------------------------------------------------------------------------------------
        The paths from `ignore_paths` and the `path_sep` parameter work exactly the same way as for
        the function `Data.get_path_id()`. See its documentation for more details."""

        def process_ignore_paths(ignore_paths: str | list[str], ) -> list[list[str]]:
            if isinstance(ignore_paths, str):
                ignore_paths = [ignore_paths]
            return [path.split(path_sep) for path in ignore_paths if path]

        def compare(
            d1: DataStructure,
            d2: DataStructure,
            ignore_paths: list[list[str]],
            current_path: list[str] = [],
        ) -> bool:
            if any(current_path == path[:len(current_path)] for path in ignore_paths):
                return True
            if type(d1) is not type(d2):
                return False
            if isinstance(d1, dict) and isinstance(d2, dict):
                if set(d1.keys()) != set(d2.keys()):
                    return False
                return all(compare(d1[key], d2[key], ignore_paths, current_path + [key]) for key in d1)
            if isinstance(d1, (list, tuple)):
                if len(d1) != len(d2):
                    return False
                return all(
                    compare(item1, item2, ignore_paths, current_path + [str(i)])
                    for i, (item1, item2) in enumerate(zip(d1, d2))
                )
            if isinstance(d1, (set, frozenset)):
                return d1 == d2
            return d1 == d2

        processed_data1 = Data.remove_comments(data1, comment_start, comment_end)
        processed_data2 = Data.remove_comments(data2, comment_start, comment_end)
        processed_ignore_paths = process_ignore_paths(ignore_paths)
        return compare(processed_data1, processed_data2, processed_ignore_paths)

    @staticmethod
    def get_path_id(
        data: DataStructure,
        value_paths: str | list[str],
        path_sep: str = "->",
        comment_start: str = ">>",
        comment_end: str = "<<",
        ignore_not_found: bool = False,
    ) -> Optional[str | list[Optional[str]]]:
        """Generates a unique ID based on the path to a specific value within a nested data structure.\n
        -------------------------------------------------------------------------------------------------
        The `data` parameter is the list, tuple, or dictionary, which the id should be generated for.\n
        -------------------------------------------------------------------------------------------------
        The param `value_path` is a sort of path (or a list of paths) to the value/s to be updated.
        In this example:
        ```python
        {
            "healthy": {
                "fruit": ["apples", "bananas", "oranges"],
                "vegetables": ["carrots", "broccoli", "celery"]
            }
        }
        ```
        ... if you want to change the value of `"apples"` to `"strawberries"`, the value path would be
        `healthy->fruit->apples` or if you don't know that the value is `"apples"` you can also use the
        index of the value, so `healthy->fruit->0`.\n
        -------------------------------------------------------------------------------------------------
        The comments marked with `comment_start` and `comment_end` will be removed, before trying to get
        the path id.\n
        -------------------------------------------------------------------------------------------------
        The `path_sep` param is the separator between the keys/indexes in the path (default is `->` just
        like in the example above).\n
        -------------------------------------------------------------------------------------------------
        If `ignore_not_found` is `True`, the function will return `None` if the value is not found
        instead of raising an error."""

        def process_path(path: str, data_obj: DataStructure) -> Optional[str]:
            keys = path.split(path_sep)
            path_ids = []
            max_id_length = 0
            for key in keys:
                if isinstance(data_obj, dict):
                    if key.isdigit():
                        if ignore_not_found:
                            return None
                        raise TypeError(f"Key '{key}' is invalid for a dict type.")
                    try:
                        idx = list(data_obj.keys()).index(key)
                        data_obj = data_obj[key]
                    except (ValueError, KeyError):
                        if ignore_not_found:
                            return None
                        raise KeyError(f"Key '{key}' not found in dict.")
                elif isinstance(data_obj, IndexIterable):
                    try:
                        idx = int(key)
                        data_obj = list(data_obj)[idx]  # CONVERT TO LIST FOR INDEXING
                    except ValueError:
                        try:
                            idx = list(data_obj).index(key)
                            data_obj = list(data_obj)[idx]
                        except ValueError:
                            if ignore_not_found:
                                return None
                            raise ValueError(f"Value '{key}' not found in '{type(data_obj).__name__}'")
                else:
                    break
                path_ids.append(str(idx))
                max_id_length = max(max_id_length, len(str(idx)))
            if not path_ids:
                return None
            return f"{max_id_length}>{''.join(id.zfill(max_id_length) for id in path_ids)}"

        data = Data.remove_comments(data, comment_start, comment_end)
        if isinstance(value_paths, str):
            return process_path(value_paths, data)
        results = [process_path(path, data) for path in value_paths]
        return results if len(results) > 1 else results[0] if results else None

    @staticmethod
    def get_value_by_path_id(data: DataStructure, path_id: str, get_key: bool = False) -> Any:
        """Retrieves the value from `data` using the provided `path_id`.\n
        -------------------------------------------------------------------------------------------------
        Input your `data` along with a `path_id` that was created before using `Data.get_path_id()`.
        If `get_key` is true and the final item is in a dict, it returns the key instead of the value.\n
        -------------------------------------------------------------------------------------------------
        The function will return the value (or key) from the path ID location, as long as the structure
        of `data` hasn't changed since creating the path ID to that value."""

        def get_nested(data: DataStructure, path: list[int], get_key: bool) -> Any:
            parent = None
            for i, idx in enumerate(path):
                if isinstance(data, dict):
                    keys = list(data.keys())
                    if i == len(path) - 1 and get_key:
                        return keys[idx]
                    parent = data
                    data = data[keys[idx]]
                elif isinstance(data, IndexIterable):
                    if i == len(path) - 1 and get_key:
                        if parent is None or not isinstance(parent, dict):
                            raise ValueError("Cannot get key from a non-dict parent")
                        return next(key for key, value in parent.items() if value is data)
                    parent = data
                    data = list(data)[idx]  # CONVERT TO LIST FOR INDEXING
                else:
                    raise TypeError(f"Unsupported type '{type(data)}' at path '{path[:i+1]}'")
            return data

        return get_nested(data, Data.__sep_path_id(path_id), get_key)

    @staticmethod
    def set_value_by_path_id(data: DataStructure, update_values: dict[str, Any]) -> DataStructure:
        """Updates the value/s from `update_values` in the `data`.\n
        --------------------------------------------------------------------------------
        Input a list, tuple or dict as `data`, along with `update_values`, which is a
        dictionary where keys are path IDs and values are the new values to insert:
           { "1>012": "new value", "1>31": ["new value 1", "new value 2"], ... }
        The path IDs should have been created using `Data.get_path_id()`.\n
        --------------------------------------------------------------------------------
        The value from path ID will be changed to the new value, as long as the
        structure of `data` hasn't changed since creating the path ID to that value."""

        def update_nested(data: DataStructure, path: list[int], value: Any) -> DataStructure:
            if len(path) == 1:
                if isinstance(data, dict):
                    keys, data = list(data.keys()), dict(data)
                    data[keys[path[0]]] = value
                elif isinstance(data, IndexIterable):
                    was_t, data = type(data), list(data)
                    data[path[0]] = value
                    data = was_t(data)
            else:
                if isinstance(data, dict):
                    keys, data = list(data.keys()), dict(data)
                    data[keys[path[0]]] = update_nested(data[keys[path[0]]], path[1:], value)
                elif isinstance(data, IndexIterable):
                    was_t, data = type(data), list(data)
                    data[path[0]] = update_nested(data[path[0]], path[1:], value)
                    data = was_t(data)
            return data

        valid_entries = [(path_id, new_val) for path_id, new_val in update_values.items()]
        if not valid_entries:
            raise ValueError(f"No valid update_values found in dictionary: {update_values}")
        for path_id, new_val in valid_entries:
            path = Data.__sep_path_id(path_id)
            data = update_nested(data, path, new_val)
        return data

    @staticmethod
    def to_str(
        data: DataStructure,
        indent: int = 4,
        compactness: int = 1,
        max_width: int = 127,
        sep: str = ", ",
        as_json: bool = False,
        _syntax_highlighting: dict[str, str] | bool = False,
    ) -> str:
        """Get nicely formatted data structure-strings.\n
        ------------------------------------------------------------------------------
        The indentation spaces-amount can be set with with `indent`.
        There are three different levels of `compactness`:
        - `0` expands everything possible
        - `1` only expands if there's other lists, tuples or dicts inside of data or,
          if the data's content is longer than `max_width`
        - `2` keeps everything collapsed (all on one line)\n
        ------------------------------------------------------------------------------
        If `as_json` is set to `True`, the output will be in valid JSON format."""
        _syntax_hl = {}
        if do_syntax_hl := _syntax_highlighting not in (None, False):
            if _syntax_highlighting is True:
                _syntax_highlighting = {}
            elif not isinstance(_syntax_highlighting, dict):
                raise TypeError(f"Expected 'syntax_highlighting' to be a dict or bool. Got: {type(_syntax_highlighting)}")
            _syntax_hl = {
                "str": (f"[{COLOR.blue}]", "[_c]"),
                "number": (f"[{COLOR.magenta}]", "[_c]"),
                "literal": (f"[{COLOR.cyan}]", "[_c]"),
                "type": (f"[i|{COLOR.lightblue}]", "[_i|_c]"),
                "punctuation": (f"[{COLOR.darkgray}]", "[_c]"),
            }
            _syntax_hl.update({
                k: (f"[{v}]", "[_]") if k in _syntax_hl and v not in ("", None) else ("", "")
                for k, v in _syntax_highlighting.items()
            })
            sep = f"{_syntax_hl['punctuation'][0]}{sep}{_syntax_hl['punctuation'][1]}"
        punct_map = {"(": ("/(", "("), **{char: char for char in "'\":)[]{}"}}
        punct = {
            k: ((f"{_syntax_hl['punctuation'][0]}{v[0]}{_syntax_hl['punctuation'][1]}" if do_syntax_hl else v[1])
                if isinstance(v, (list, tuple)) else
                (f"{_syntax_hl['punctuation'][0]}{v}{_syntax_hl['punctuation'][1]}" if do_syntax_hl else v))
            for k, v in punct_map.items()
        }

        def format_value(value: Any, current_indent: Optional[int] = None) -> str:
            if current_indent is not None and isinstance(value, dict):
                return format_dict(value, current_indent + indent)
            elif current_indent is not None and hasattr(value, "__dict__"):
                return format_dict(value.__dict__, current_indent + indent)
            elif current_indent is not None and isinstance(value, IndexIterable):
                return format_sequence(value, current_indent + indent)
            elif current_indent is not None and isinstance(value, (bytes, bytearray)):
                obj_dict = Data.serialize_bytes(value)
                return (
                    format_dict(obj_dict, current_indent + indent) if as_json else (
                        f"{_syntax_hl['type'][0]}{(k := next(iter(obj_dict)))}{_syntax_hl['type'][1]}"
                        + format_sequence((obj_dict[k], obj_dict["encoding"]), current_indent + indent) if do_syntax_hl else
                        (k := next(iter(obj_dict)))
                        + format_sequence((obj_dict[k], obj_dict["encoding"]), current_indent + indent)
                    )
                )
            elif isinstance(value, bool):
                val = str(value).lower() if as_json else str(value)
                return f"{_syntax_hl['literal'][0]}{val}{_syntax_hl['literal'][1]}" if do_syntax_hl else val
            elif isinstance(value, (int, float)):
                val = "null" if as_json and (_math.isinf(value) or _math.isnan(value)) else str(value)
                return f"{_syntax_hl['number'][0]}{val}{_syntax_hl['number'][1]}" if do_syntax_hl else val
            elif current_indent is not None and isinstance(value, complex):
                return (
                    format_value(str(value).strip("()")) if as_json else (
                        f"{_syntax_hl['type'][0]}complex{_syntax_hl['type'][1]}"
                        + format_sequence((value.real, value.imag), current_indent + indent)
                        if do_syntax_hl else f"complex{format_sequence((value.real, value.imag), current_indent + indent)}"
                    )
                )
            elif value is None:
                val = "null" if as_json else "None"
                return f"{_syntax_hl['literal'][0]}{val}{_syntax_hl['literal'][1]}" if do_syntax_hl else val
            else:
                return ((
                    punct['"'] + _syntax_hl["str"][0] + String.escape(str(value), '"') + _syntax_hl["str"][1]
                    + punct['"'] if do_syntax_hl else punct['"'] + String.escape(str(value), '"') + punct['"']
                ) if as_json else (
                    punct["'"] + _syntax_hl["str"][0] + String.escape(str(value), "'") + _syntax_hl["str"][1]
                    + punct["'"] if do_syntax_hl else punct["'"] + String.escape(str(value), "'") + punct["'"]
                ))

        def should_expand(seq: IndexIterable) -> bool:
            if compactness == 0:
                return True
            if compactness == 2:
                return False
            complex_types = (list, tuple, dict, set, frozenset) + ((bytes, bytearray) if as_json else ())
            complex_items = sum(1 for item in seq if isinstance(item, complex_types))
            return (
                complex_items > 1 or (complex_items == 1 and len(seq) > 1) or Data.chars_count(seq) +
                (len(seq) * len(sep)) > max_width
            )

        def format_dict(d: dict, current_indent: int) -> str:
            if not d or compactness == 2:
                return (
                    punct["{"]
                    + sep.join(f"{format_value(k)}{punct[':']} {format_value(v, current_indent)}"
                               for k, v in d.items()) + punct["}"]
                )
            if not should_expand(list(d.values())):
                return (
                    punct["{"]
                    + sep.join(f"{format_value(k)}{punct[':']} {format_value(v, current_indent)}"
                               for k, v in d.items()) + punct["}"]
                )
            items = []
            for k, val in d.items():
                formatted_value = format_value(val, current_indent)
                items.append(f"{' ' * (current_indent + indent)}{format_value(k)}{punct[':']} {formatted_value}")
            return punct["{"] + "\n" + f"{sep}\n".join(items) + f"\n{' ' * current_indent}" + punct["}"]

        def format_sequence(seq, current_indent: int) -> str:
            if as_json:
                seq = list(seq)
            if not seq or compactness == 2:
                return (
                    punct["["] + sep.join(format_value(item, current_indent)
                                          for item in seq) + punct["]"] if isinstance(seq, list) else punct["("]
                    + sep.join(format_value(item, current_indent) for item in seq) + punct[")"]
                )
            if not should_expand(seq):
                return (
                    punct["["] + sep.join(format_value(item, current_indent)
                                          for item in seq) + punct["]"] if isinstance(seq, list) else punct["("]
                    + sep.join(format_value(item, current_indent) for item in seq) + punct[")"]
                )
            items = [format_value(item, current_indent) for item in seq]
            formatted_items = f"{sep}\n".join(f'{" " * (current_indent + indent)}{item}' for item in items)
            if isinstance(seq, list):
                return f"{punct['[']}\n{formatted_items}\n{' ' * current_indent}{punct[']']}"
            else:
                return f"{punct['(']}\n{formatted_items}\n{' ' * current_indent}{punct[')']}"

        return _re.sub(r"\s+(?=\n)", "", format_dict(data, 0) if isinstance(data, dict) else format_sequence(data, 0))

    @staticmethod
    def print(
        data: DataStructure,
        indent: int = 4,
        compactness: int = 1,
        max_width: int = 127,
        sep: str = ", ",
        end: str = "\n",
        as_json: bool = False,
        syntax_highlighting: dict[str, str] | bool = {},
    ) -> None:
        """Print nicely formatted data structures.\n
        ------------------------------------------------------------------------------
        The indentation spaces-amount can be set with with `indent`.
        There are three different levels of `compactness`:
        - `0` expands everything possible
        - `1` only expands if there's other lists, tuples or dicts inside of data or,
          if the data's content is longer than `max_width`
        - `2` keeps everything collapsed (all on one line)\n
        ------------------------------------------------------------------------------
        If `as_json` is set to `True`, the output will be in valid JSON format.\n
        ------------------------------------------------------------------------------
        The `syntax_highlighting` parameter is a dictionary with 5 keys for each part
        of the data. The key's values are the formatting codes to apply to this data
        part. The formatting can be changed by simply adding the key with the new
        value inside the `syntax_highlighting` dictionary.\n
        The keys with their default values are:
        - `str: COLOR.blue`
        - `number: COLOR.magenta`
        - `literal: COLOR.cyan`
        - `type: "i|" + COLOR.lightblue`
        - `punctuation: COLOR.darkgray`\n
        For no syntax highlighting, set `syntax_highlighting` to `False` or `None`.\n
        ------------------------------------------------------------------------------
        For more detailed information about formatting codes, see `xx_format_codes`
        module documentation."""
        FormatCodes.print(
            Data.to_str(data, indent, compactness, max_width, sep, as_json, syntax_highlighting),
            end=end,
        )

    @staticmethod
    def __sep_path_id(path_id: str) -> list[int]:
        if path_id.count(">") != 1:
            raise ValueError(f"Invalid path ID: {path_id}")
        id_part_len = int(path_id.split(">")[0])
        path_ids_str = path_id.split(">")[1]
        return [int(path_ids_str[i:i + id_part_len]) for i in range(0, len(path_ids_str), id_part_len)]



__version__ = "1.7.0"
__author__ = "XulbuX"
__email__ = "xulbux.real@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024 XulbuX"
__url__ = "https://github.com/XulbuX/PythonLibraryXulbuX"
__description__ = "A library which includes a lot of really helpful functions."
__all__ = [
    "_consts_", "xx_code", "xx_color", "xx_console", "xx_data", "xx_env_path", "xx_file", "xx_format_codes", "xx_json",
    "xx_path", "xx_regex", "xx_string", "xx_system"
]

from ._consts_ import *
from .xx_code import *
from .xx_color import *
from .xx_console import *
from .xx_data import *
from .xx_env_path import *
from .xx_file import *
from .xx_format_codes import *
from .xx_json import *
from .xx_path import *
from .xx_regex import *
from .xx_string import *
from .xx_system import *



"""
Functions for logging and other small actions within the console.\n
----------------------------------------------------------------------------------------------------------
You can also use special formatting codes directly inside the log message to change their appearance.
For more detailed information about formatting codes, see the the `xx_format_codes` module documentation.
"""

from ._consts_ import COLOR, CHARS
from .xx_format_codes import FormatCodes, _COMPILED
from .xx_string import String
from .xx_color import Color, Rgba, Hexa

from prompt_toolkit.key_binding.key_bindings import KeyBindings
from typing import Optional, Any
import prompt_toolkit as _prompt_toolkit
import pyperclip as _pyperclip
import keyboard as _keyboard
import getpass as _getpass
import shutil as _shutil
import mouse as _mouse
import sys as _sys
import os as _os


class _ConsoleWidth:

    def __get__(self, obj, owner=None):
        return _os.get_terminal_size().columns


class _ConsoleHeight:

    def __get__(self, obj, owner=None):
        return _os.get_terminal_size().lines


class _ConsoleSize:

    def __get__(self, obj, owner=None):
        size = _os.get_terminal_size()
        return (size.columns, size.lines)


class _ConsoleUser:

    def __get__(self, obj, owner=None):
        return _os.getenv("USER") or _os.getenv("USERNAME") or _getpass.getuser()


class ArgResult:
    """Exists: if the argument was found or not\n
    Value: the value from behind the found argument"""

    def __init__(self, exists: bool, value: Any):
        self.exists = exists
        self.value = value

    def __bool__(self):
        return self.exists


class Args:
    """Stores found command arguments under their aliases with their results."""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if not key.isidentifier():
                raise TypeError(f"Argument alias '{key}' is invalid. It must be a valid Python variable name.")
            setattr(self, key, ArgResult(**value))

    def __len__(self):
        return len(vars(self))

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        if isinstance(key, int):
            return list(self.__iter__())[key]
        return getattr(self, key)

    def __iter__(self):
        for key, value in vars(self).items():
            yield (key, {"exists": value.exists, "value": value.value})

    def dict(self) -> dict[str, dict[str, Any]]:
        """Returns the arguments as a dictionary."""
        return {k: {"exists": v.exists, "value": v.value} for k, v in vars(self).items()}

    def keys(self):
        """Returns the argument aliases as `dict_keys([...])`."""
        return vars(self).keys()

    def values(self):
        """Returns the argument results as `dict_values([...])`."""
        return vars(self).values()

    def items(self):
        """Yields tuples of `(alias, {'exists': bool, 'value': Any})`."""
        for key, value in self.__iter__():
            yield (key, value)


class Console:

    w: int = _ConsoleWidth()  # type: ignore[assignment]
    """The width of the console in characters."""
    h: int = _ConsoleHeight()  # type: ignore[assignment]
    """The height of the console in lines."""
    wh: tuple[int, int] = _ConsoleSize()  # type: ignore[assignment]
    """A tuple with the width and height of
    the console in characters and lines."""
    usr: str = _ConsoleUser()  # type: ignore[assignment]
    """The name of the current user."""

    @staticmethod
    def get_args(
        find_args: dict[str, list[str] | tuple[str, ...] | dict[str, list[str] | tuple[str, ...] | Any]],
        allow_spaces: bool = False
    ) -> Args:
        """Will search for the specified arguments in the command line
        arguments and return the results as a special `Args` object.\n
        ----------------------------------------------------------------
        The `find_args` dictionary can have the following structures for each alias:
        1. Simple list/tuple of flags (when no default value is needed):
           ```python
           "alias_name": ["-f", "--flag"]
           ```
        2. Dictionary with 'flags' and optional 'default':
           ```python
           "alias_name": {
               "flags": ["-f", "--flag"],
               "default": "some_value"  # Optional
           }
           ```
        Example `find_args`:
        ```python
        find_args={
            "arg1": { # With default
                "flags": ["-a1", "--arg1"],
                "default": "default_val"
            },
            "arg2": ("-a2", "--arg2"), # Without default (original format)
            "arg3": ["-a3"],          # Without default (list format)
            "arg4": { # Flag with default True
                "flags": ["-f"],
                "default": True
            }
        }
        ```
        If the script is called via the command line:\n
        `python script.py -a1 "value1" --arg2 -f`\n
        ...it would return an `Args` object where:
        - `args.arg1.exists` is `True`, `args.arg1.value` is `"value1"`
        - `args.arg2.exists` is `True`, `args.arg2.value` is `True` (flag present without value)
        - `args.arg3.exists` is `False`, `args.arg3.value` is `None` (not present, no default)
        - `args.arg4.exists` is `True`, `args.arg4.value` is `True` (flag present, overrides default)
        - If an arg defined in `find_args` is *not* present in the command line:
            - `exists` will be `False`
            - `value` will be the specified `default` value, or `None` if no default was specified.\n
        ----------------------------------------------------------------
        Normally if `allow_spaces` is false, it will take a space as
        the end of an args value. If it is true, it will take spaces as
        part of the value until the next arg is found.
        (Multiple spaces will become one space in the value.)"""
        args = _sys.argv[1:]
        args_len = len(args)
        arg_lookup = {}
        results = {}
        for alias, config in find_args.items():
            flags = None
            default_value = None
            if isinstance(config, (list, tuple)):
                flags = config
            elif isinstance(config, dict):
                if "flags" not in config:
                    raise ValueError(f"Invalid configuration for alias '{alias}'. Dictionary must contain a 'flags' key.")
                flags = config["flags"]
                default_value = config.get("default")
                if not isinstance(flags, (list, tuple)):
                    raise ValueError(f"Invalid 'flags' for alias '{alias}'. Must be a list or tuple.")
            else:
                raise TypeError(f"Invalid configuration type for alias '{alias}'. Must be a list, tuple, or dict.")
            results[alias] = {"exists": False, "value": default_value}
            for flag in flags:
                if flag in arg_lookup:
                    raise ValueError(
                        f"Duplicate flag '{flag}' found. It's assigned to both '{arg_lookup[flag]}' and '{alias}'."
                    )
                arg_lookup[flag] = alias
        i = 0
        while i < args_len:
            arg = args[i]
            alias = arg_lookup.get(arg)
            if alias:
                results[alias]["exists"] = True
                value_found_after_flag = False
                if i + 1 < args_len and not args[i + 1].startswith("-"):
                    if not allow_spaces:
                        results[alias]["value"] = String.to_type(args[i + 1])
                        i += 1
                        value_found_after_flag = True
                    else:
                        value_parts = []
                        j = i + 1
                        while j < args_len and not args[j].startswith("-"):
                            value_parts.append(args[j])
                            j += 1
                        if value_parts:
                            results[alias]["value"] = String.to_type(" ".join(value_parts))
                            i = j - 1
                            value_found_after_flag = True
                if not value_found_after_flag:
                    results[alias]["value"] = True
            i += 1
        return Args(**results)

    @staticmethod
    def pause_exit(
        pause: bool = False,
        exit: bool = False,
        prompt: object = "",
        exit_code: int = 0,
        reset_ansi: bool = False,
    ) -> None:
        """Will print the `last_prompt` and then pause the program if `pause` is set
        to `True` and after the pause, exit the program if `exit` is set to `True`."""
        print(prompt, end="", flush=True)
        if reset_ansi:
            FormatCodes.print("[_]", end="")
        if pause:
            _keyboard.read_event()
        if exit:
            _sys.exit(exit_code)

    @staticmethod
    def cls() -> None:
        """Will clear the console in addition to completely resetting the ANSI formats."""
        if _shutil.which("cls"):
            _os.system("cls")
        elif _shutil.which("clear"):
            _os.system("clear")
        print("\033[0m", end="", flush=True)

    @staticmethod
    def log(
        title: Optional[str] = None,
        prompt: object = "",
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Optional[Rgba | Hexa] = None,
        default_color: Optional[Rgba | Hexa] = None,
        _console_tabsize: int = 8,
    ) -> None:
        """Will print a formatted log message:
        - `title` -⠀the title of the log message (e.g. `DEBUG`, `WARN`, `FAIL`, etc.)
        - `prompt` -⠀the log message
        - `format_linebreaks` -⠀whether to format (indent after) the line breaks or not
        - `start` -⠀something to print before the log is printed
        - `end` -⠀something to print after the log is printed (e.g. `\\n`)
        - `title_bg_color` -⠀the background color of the `title`
        - `default_color` -⠀the default text color of the `prompt`
        - `_console_tabsize` -⠀the tab size of the console (default is 8)\n
        -----------------------------------------------------------------------------------
        The log message can be formatted with special formatting codes. For more detailed
        information about formatting codes, see `xx_format_codes` module documentation."""
        title = "" if title is None else title.strip().upper()
        title_len, tab_len = len(title) + 4, _console_tabsize - ((len(title) + 4) % _console_tabsize)
        title_color = "_color" if not title_bg_color else Color.text_color_for_on_bg(title_bg_color)
        if format_linebreaks:
            clean_prompt, removals = FormatCodes.remove_formatting(str(prompt), get_removals=True, _ignore_linebreaks=True)
            prompt_lst = (String.split_count(l, Console.w - (title_len + tab_len)) for l in str(clean_prompt).splitlines())
            prompt_lst = (
                item for lst in prompt_lst for item in ([""] if lst == [] else (lst if isinstance(lst, list) else [lst]))
            )
            prompt = f"\n{' ' * title_len}\t".join(
                Console.__add_back_removed_parts(list(prompt_lst), removals)  # type: ignore[assignment]
            )
        else:
            prompt = str(prompt)
        if title == "":
            FormatCodes.print(
                f'{start}  {f"[{default_color}]" if default_color else ""}{str(prompt)}[_]',
                default_color=default_color,
                end=end,
            )
        else:
            FormatCodes.print(
                f'{start}  [bold][{title_color}]{f"[BG:{title_bg_color}]" if title_bg_color else ""} {title} [_]'
                + f'\t{f"[{default_color}]" if default_color else ""}{prompt}[_]',
                default_color=default_color,
                end=end,
            )

    @staticmethod
    def __add_back_removed_parts(split_string: list[str], removals: tuple[tuple[int, str], ...]) -> list[str]:
        """Adds back the removed parts into the split string parts at their original positions."""
        lengths, cumulative_pos = [len(s) for s in split_string], [0]
        for length in lengths:
            cumulative_pos.append(cumulative_pos[-1] + length)
        result, offset_adjusts = split_string.copy(), [0] * len(split_string)
        last_idx, total_length = len(split_string) - 1, cumulative_pos[-1]

        def find_string_part(pos: int) -> int:
            left, right = 0, len(cumulative_pos) - 1
            while left < right:
                mid = (left + right) // 2
                if cumulative_pos[mid] <= pos < cumulative_pos[mid + 1]:
                    return mid
                elif pos < cumulative_pos[mid]:
                    right = mid
                else:
                    left = mid + 1
            return left

        for pos, removal in removals:
            if pos >= total_length:
                result[last_idx] = result[last_idx] + removal
                continue
            i = find_string_part(pos)
            adjusted_pos = (pos - cumulative_pos[i]) + offset_adjusts[i]
            parts = [result[i][:adjusted_pos], removal, result[i][adjusted_pos:]]
            result[i] = ''.join(parts)
            offset_adjusts[i] += len(removal)
        return result

    @staticmethod
    def debug(
        prompt: object = "Point in program reached.",
        active: bool = True,
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Rgba | Hexa = COLOR.yellow,
        default_color: Rgba | Hexa = COLOR.text,
        pause: bool = False,
        exit: bool = False,
    ) -> None:
        """A preset for `log()`: `DEBUG` log message with the options to pause
        at the message and exit the program after the message was printed.
        If `active` is false, no debug message will be printed."""
        if active:
            Console.log("DEBUG", prompt, format_linebreaks, start, end, title_bg_color, default_color)
            Console.pause_exit(pause, exit)

    @staticmethod
    def info(
        prompt: object = "Program running.",
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Rgba | Hexa = COLOR.blue,
        default_color: Rgba | Hexa = COLOR.text,
        pause: bool = False,
        exit: bool = False,
    ) -> None:
        """A preset for `log()`: `INFO` log message with the options to pause
        at the message and exit the program after the message was printed."""
        Console.log("INFO", prompt, format_linebreaks, start, end, title_bg_color, default_color)
        Console.pause_exit(pause, exit)

    @staticmethod
    def done(
        prompt: object = "Program finished.",
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Rgba | Hexa = COLOR.teal,
        default_color: Rgba | Hexa = COLOR.text,
        pause: bool = False,
        exit: bool = False,
    ) -> None:
        """A preset for `log()`: `DONE` log message with the options to pause
        at the message and exit the program after the message was printed."""
        Console.log("DONE", prompt, format_linebreaks, start, end, title_bg_color, default_color)
        Console.pause_exit(pause, exit)

    @staticmethod
    def warn(
        prompt: object = "Important message.",
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Rgba | Hexa = COLOR.orange,
        default_color: Rgba | Hexa = COLOR.text,
        pause: bool = False,
        exit: bool = False,
    ) -> None:
        """A preset for `log()`: `WARN` log message with the options to pause
        at the message and exit the program after the message was printed."""
        Console.log("WARN", prompt, format_linebreaks, start, end, title_bg_color, default_color)
        Console.pause_exit(pause, exit)

    @staticmethod
    def fail(
        prompt: object = "Program error.",
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Rgba | Hexa = COLOR.red,
        default_color: Rgba | Hexa = COLOR.text,
        pause: bool = False,
        exit: bool = True,
        reset_ansi=True,
    ) -> None:
        """A preset for `log()`: `FAIL` log message with the options to pause
        at the message and exit the program after the message was printed."""
        Console.log("FAIL", prompt, format_linebreaks, start, end, title_bg_color, default_color)
        Console.pause_exit(pause, exit, reset_ansi=reset_ansi)

    @staticmethod
    def exit(
        prompt: object = "Program ended.",
        format_linebreaks: bool = True,
        start: str = "",
        end: str = "\n",
        title_bg_color: Rgba | Hexa = COLOR.magenta,
        default_color: Rgba | Hexa = COLOR.text,
        pause: bool = False,
        exit: bool = True,
        reset_ansi=True,
    ) -> None:
        """A preset for `log()`: `EXIT` log message with the options to pause
        at the message and exit the program after the message was printed."""
        Console.log("EXIT", prompt, format_linebreaks, start, end, title_bg_color, default_color)
        Console.pause_exit(pause, exit, reset_ansi=reset_ansi)

    @staticmethod
    def log_box(
        *values: object,
        start: str = "",
        end: str = "\n",
        box_bg_color: str | Rgba | Hexa = "green",
        default_color: Rgba | Hexa = "#000",
        w_padding: int = 2,
        w_full: bool = False,
    ) -> None:
        """Will print a box, containing a formatted log message:
        - `*values` -⠀the box content (each value is on a new line)
        - `start` -⠀something to print before the log box is printed
        - `end` -⠀something to print after the log box is printed (e.g. `\\n`)
        - `box_bg_color` -⠀the background color of the box
        - `default_color` -⠀the default text color of the `*values`
        - `w_padding` -⠀the horizontal padding (in chars) to the box content
        - `w_full` -⠀whether to make the box be the full console width or not\n
        -----------------------------------------------------------------------------------
        The box content can be formatted with special formatting codes. For more detailed
        information about formatting codes, see `xx_format_codes` module documentation."""
        lines = [line.rstrip() for val in values for line in str(val).splitlines()]
        unfmt_lines = [FormatCodes.remove_formatting(line) for line in lines]
        max_line_len = max(len(line) for line in unfmt_lines)
        pad_w_full = (Console.w - (max_line_len + (2 * w_padding))) if w_full else 0
        lines = [
            f"[bg:{box_bg_color}]{' ' * w_padding}{line}" + " " *
            ((w_padding + max_line_len - len(unfmt)) + pad_w_full) + "[_bg]" for line, unfmt in zip(lines, unfmt_lines)
        ]
        pady = " " * (Console.w if w_full else max_line_len + (2 * w_padding))
        FormatCodes.print(
            f"{start}[bg:{box_bg_color}]{pady}[_bg]\n"
            + _COMPILED["formatting"].sub(lambda m: f"{m.group(0)}[bg:{box_bg_color}]", "\n".join(lines))
            + f"\n[bg:{box_bg_color}]{pady}[_bg]",
            default_color=default_color,
            sep="\n",
            end=end,
        )

    @staticmethod
    def confirm(
        prompt: object = "Do you want to continue?",
        start="",
        end="\n",
        default_color: Rgba | Hexa = COLOR.cyan,
        default_is_yes: bool = True,
    ) -> bool:
        """Ask a yes/no question.\n
        ---------------------------------------------------------------------------------------
        The prompt can be formatted with special formatting codes. For more detailed
        information about formatting codes, see the `xx_format_codes` module documentation."""
        confirmed = input(
            FormatCodes.to_ansi(
                f'{start}  {str(prompt)} [_|dim](({"Y" if default_is_yes else "y"}/{"n" if default_is_yes else "N"}):  )',
                default_color,
            )
        ).strip().lower() in (("", "y", "yes") if default_is_yes else ("y", "yes"))
        if end:
            Console.log("", end, end="")
        return confirmed

    @staticmethod
    def multiline_input(
        prompt: object = "",
        start="",
        end="\n",
        default_color: Rgba | Hexa = COLOR.cyan,
        show_keybindings=True,
        input_prefix=" ⤷ ",
        reset_ansi=True,
    ) -> str:
        """An input where users can input (and paste) text over multiple lines.\n
        -----------------------------------------------------------------------------------
        - `prompt` -⠀the input prompt
        - `start` -⠀something to print before the input
        - `end` -⠀something to print after the input (e.g. `\\n`)
        - `default_color` -⠀the default text color of the `prompt`
        - `show_keybindings` -⠀whether to show the special keybindings or not
        - `input_prefix` -⠀the prefix of the input line
        - `reset_ansi` -⠀whether to reset the ANSI codes after the input or not
        -----------------------------------------------------------------------------------
        The input prompt can be formatted with special formatting codes. For more detailed
        information about formatting codes, see `xx_format_codes` module documentation."""
        kb = KeyBindings()

        @kb.add("c-d", eager=True)  # CTRL+D
        def _(event):
            event.app.exit(result=event.app.current_buffer.document.text)

        FormatCodes.print(start + str(prompt), default_color=default_color)
        if show_keybindings:
            FormatCodes.print("[dim][[b](CTRL+D)[dim] : end of input][_dim]")
        input_string = _prompt_toolkit.prompt(input_prefix, multiline=True, wrap_lines=True, key_bindings=kb)
        FormatCodes.print("[_]" if reset_ansi else "", end=end[1:] if end.startswith("\n") else end)
        return input_string

    @staticmethod
    def restricted_input(
        prompt: object = "",
        start="",
        end="\n",
        default_color: Rgba | Hexa = COLOR.cyan,
        allowed_chars: str = CHARS.all,  # type: ignore[assignment]
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
        mask_char: Optional[str] = None,
        reset_ansi: bool = True,
    ) -> Optional[str]:
        """Acts like a standard Python `input()` with the advantage, that you can specify:
        - what text characters the user is allowed to type and
        - the minimum and/or maximum length of the users input
        - optional mask character (hide user input, e.g. for passwords)
        - reset the ANSI formatting codes after the user continues\n
        ---------------------------------------------------------------------------------------
        The input can be formatted with special formatting codes. For more detailed
        information about formatting codes, see the `xx_format_codes` module documentation."""
        FormatCodes.print(start + str(prompt), default_color=default_color, end="")
        result = ""
        select_all = False
        last_line_count = 1
        last_console_width = 0

        def update_display(console_width: int) -> None:
            nonlocal last_line_count, last_console_width
            lines = String.split_count(str(prompt) + (mask_char * len(result) if mask_char else result), console_width)
            line_count = len(lines)
            if (line_count > 1 or line_count < last_line_count) and not last_line_count == 1:
                if last_console_width > console_width:
                    line_count *= 2
                for _ in range(line_count if line_count < last_line_count and not line_count > last_line_count else (
                        line_count - 2 if line_count > last_line_count else line_count - 1)):
                    _sys.stdout.write("\033[2K\r\033[A")
            prompt_len = len(str(prompt)) if prompt else 0
            prompt_str = lines[0][:prompt_len]
            input_str = (
                lines[0][prompt_len:] if len(lines) == 1 else "\n".join([lines[0][prompt_len:]] + lines[1:])
            )  # SEPARATE THE PROMPT AND THE INPUT
            _sys.stdout.write(
                "\033[2K\r" + FormatCodes.to_ansi(prompt_str) + ("\033[7m" if select_all else "") + input_str + "\033[27m"
            )
            last_line_count, last_console_width = line_count, console_width

        def handle_enter():
            if min_len is not None and len(result) < min_len:
                return False
            FormatCodes.print(f"[_]{end}" if reset_ansi else end, default_color=default_color)
            return True

        def handle_backspace_delete():
            nonlocal result, select_all
            if select_all:
                result, select_all = "", False
            elif result and event.name == "backspace":
                result = result[:-1]
            update_display(Console.w)

        def handle_paste():
            nonlocal result, select_all
            if select_all:
                result, select_all = "", False
            filtered_text = "".join(char for char in _pyperclip.paste() if allowed_chars == CHARS.all or char in allowed_chars)
            if max_len is None or len(result) + len(filtered_text) <= max_len:
                result += filtered_text
                update_display(Console.w)

        def handle_select_all():
            nonlocal select_all
            select_all = True
            update_display(Console.w)

        def handle_character_input():
            nonlocal result
            if event.name is not None and ((allowed_chars == CHARS.all or event.name in allowed_chars) and
                                           (max_len is None or len(result) < max_len)):
                result += event.name
                update_display(Console.w)

        while True:
            event = _keyboard.read_event()
            if event.event_type == "down":
                if event.name == "enter" and handle_enter():
                    return result.rstrip("\n")
                elif event.name in ("backspace", "delete", "entf"):
                    handle_backspace_delete()
                elif (event.name == "v" and _keyboard.is_pressed("ctrl")) or _mouse.is_pressed("right"):
                    handle_paste()
                elif event.name == "a" and _keyboard.is_pressed("ctrl"):
                    handle_select_all()
                elif event.name == "c" and _keyboard.is_pressed("ctrl"):
                    raise KeyboardInterrupt
                elif event.name == "esc":
                    return None
                elif event.name == "space":
                    handle_character_input()
                elif event.name is not None and len(event.name) == 1:
                    handle_character_input()
                else:
                    select_all = False
                    update_display(Console.w)

    @staticmethod
    def pwd_input(
        prompt: object = "Password: ",
        start="",
        end="\n",
        default_color: Rgba | Hexa = COLOR.cyan,
        allowed_chars: str = CHARS.standard_ascii,
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
        reset_ansi: bool = True,
    ) -> Optional[str]:
        """Password input (preset for `Console.restricted_input()`)
        that always masks the entered characters with asterisks."""
        return Console.restricted_input(
            prompt=prompt,
            start=start,
            end=end,
            default_color=default_color,
            allowed_chars=allowed_chars,
            min_len=min_len,
            max_len=max_len,
            mask_char="*",
            reset_ansi=reset_ansi,
        )
